cdef extern from "JavaScriptCore/JSObjectRef.h":
    enum :
        kJSPropertyAttributeNone,kJSPropertyAttributeReadOnly,
        kJSPropertyAttributeDontEnum,kJSPropertyAttributeDontDelete
    ctypedef unsigned JSPropertyAttributes
    bool JSObjectHasProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName)
    JSValueRef JSObjectGetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
    void JSObjectSetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef value, JSPropertyAttributes attributes, JSValueRef* exception)
    void JSObjectDeleteProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
    JSValueRef JSObjectGetPropertyAtIndex(JSContextRef ctx, JSObjectRef object, unsigned propertyIndex, JSValueRef* exception)
    void JSObjectSetPropertyAtIndex(JSContextRef ctx, JSObjectRef object, unsigned propertyIndex, JSValueRef value, JSValueRef* exception)
    void* JSObjectGetPrivate(JSObjectRef object)
    bool JSObjectSetPrivate(JSObjectRef object, void* data)
    bool JSObjectIsFunction(JSContextRef ctx, JSObjectRef object)
    JSValueRef JSObjectCallAsFunction(JSContextRef ctx, JSObjectRef object, JSObjectRef thisObject, size_t argumentCount,JSValueRef arguments[], JSValueRef* exception)
    JSObjectIsConstructor(JSContextRef ctx, JSObjectRef object)
    JSPropertyNameArrayRef JSObjectCopyPropertyNames(JSContextRef ctx, JSObjectRef object)
    JSPropertyNameArrayRef JSPropertyNameArrayRetain(JSPropertyNameArrayRef array)
    void JSPropertyNameArrayRelease(JSPropertyNameArrayRef array)
    size_t JSPropertyNameArrayGetCount(JSPropertyNameArrayRef array)
    JSStringRef JSPropertyNameArrayGetNameAtIndex(JSPropertyNameArrayRef array, size_t index)
    void JSPropertyNameAccumulatorAddName(JSPropertyNameAccumulatorRef accumulator, JSStringRef propertyName)


cdef extern from "stdlib.h":
    ctypedef unsigned long size_t
    void free(void *ptr)
    void *malloc(size_t size)
    void *realloc(void *ptr, size_t size)
    size_t strlen(char *s)
    char *strcpy(char *dest, char *src)



cdef extern from "JavaScriptCore/JSValueRef.h":
    cdef enum JSType :
         kJSTypeUndefined,kJSTypeNull,kJSTypeBoolean,kJSTypeNumber,kJSTypeString,kJSTypeObject
    JSType JSValueGetType(JSContextRef ctx, JSValueRef value)
    bool JSValueToBoolean(JSContextRef ctx, JSValueRef value)
    double JSValueToNumber(JSContextRef ctx, JSValueRef value, JSValueRef* exception)
    JSStringRef JSValueToStringCopy(JSContextRef ctx, JSValueRef value, JSValueRef* exception)
    JSObjectRef JSValueToObject(JSContextRef ctx, JSValueRef value, JSValueRef* exception)
    
    JSValueRef JSValueMakeUndefined(JSContextRef ctx)
    JSValueRef JSValueMakeNull(JSContextRef ctx)
    JSValueRef JSValueMakeBoolean(JSContextRef ctx, bool boolean)
    JSValueRef JSValueMakeNumber(JSContextRef ctx, double number)
    JSValueRef JSValueMakeString(JSContextRef ctx, JSStringRef string)
    void JSValueProtect(JSContextRef ctx, JSValueRef value)
    void JSValueUnprotect(JSContextRef ctx, JSValueRef value)


cdef extern from "Python.h":
    ctypedef unsigned short Py_UNICODE

cdef extern from "JavaScriptCore/JSStringRef.h":
    ctypedef Py_UNICODE JSChar
    void JSStringRelease(JSStringRef string)
    JSStringRef JSStringCreateWithUTF8CString(char* string)
    JSStringRef JSStringCreateWithCharacters(JSChar* chars, size_t numChars)
    size_t JSStringGetLength(JSStringRef string)
    JSChar* JSStringGetCharactersPtr(JSStringRef string)
    size_t JSStringGetMaximumUTF8CStringSize(JSStringRef string)
    size_t JSStringGetUTF8CString(JSStringRef string, char* buffer, size_t bufferSize)



"""
Copyright (C) 2009 john paul janecek, Free Beer (MIT) Licensing
Copyright (C) 2010 Jiahua Huang <jhuangjiahua@gmail.com>, MIT Licensing for this file

Comment by john paul janecek:
------------------
test file for cython wrapper for JSContextRef in pywebkitgtk
This is the documentation :)
you need cython to make it
So you can call Javascript functions etc from python
Made by john paul janecek
Free Beer copyright, do what the heck you want with it, just give me credit
Also do not blame me if your things blow up
if you need to contact me, i might answer back :) I am lazy when it comes to making fixes
unless I actually am using library myself :)

my email
import binascii
binascii.a2b_base64('anBqYW5lY2VrQGdtYWlsLmNvbQ==\n')
------------------

"""

import gobject
import gtk
import pango
import webkit
import jswebkit
import signal

gobject.threads_init()

class WebView(webkit.WebView):
    def __init__(self):
        webkit.WebView.__init__(self)
        self.mainWindow = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.mainWindow.connect("delete_event", lambda *x: gtk.main_quit ())
        self.mainWindow.set_size_request(800, 600)
        self.scrolledWindow = gtk.ScrolledWindow()
        self.mainWindow.add(self.scrolledWindow)
        self.scrolledWindow.add(self)
        settings = self.get_settings()
        settings.set_property("auto-load-images",False)
        settings.set_property("enable-plugins",False)
        self.connect("load-finished",self.load_finished_cb)
        
    def show_all(self):
        self.mainWindow.show_all()

    def load_finished_cb(self,view,frame):
        print "load_finished"
        ctx = jswebkit.JSContext(self.get_main_frame().get_global_context())
        window = ctx.EvaluateScript("window")
        #window.alert(None, "window")
        #window.foo = "bar"
        #print ctx.EvaluateScript("window.foo")
        document = ctx.EvaluateScript("document")
        #print "Title : ",document.title
        #form = document.forms[0]
        #print form.action
        #form.elements[1].value = "this is me"
        #form.elements[2].click(form.elements[2])
        atags = document.getElementsByTagName(document, "a")
        print atags.getPropertyNames()
        for a in atags :
            print a.href
        
    def start(self):
        self.open("http://www.google.com")
        
try:
    view = WebView()
    view.show_all()
    view.start()
    gtk.main()
except KeyboardInterrupt:
    gtk.main_quit()




cdef extern from "Python.h":
    ctypedef unsigned short Py_UNICODE
    char* PyCObject_GetDesc(object self)
    void* PyCObject_AsVoidPtr(object self)
    char* PyString_AsString(object o)
    object PyUnicode_DecodeUTF16(Py_UNICODE *u, Py_ssize_t size, char *errors, int byteorder)


"""
setup file for cython wrapper for JSContextRef in pywebkitgtk
 
Copyright (C) 2009 john paul janecek, Free Beer (MIT) Licensing
Copyright (C) 2010 Aron Xu <happyaron.xu@gmail.com>, MIT Licensing for this file

Comment by john paul janecek:
------------------
setup file for cython wrapper for JSContextRef in pywebkitgtk
you need cython to make it
So you can call Javascript functions etc from python
Made by john paul janecek
Free Beer copyright, do what the heck you want with it, just give me credit
Also do not blame me if your things blow up
if you need to contact me, i might answer back :) I am lazy when it comes to making fixes
unless I actually am using library myself :)

my email
import binascii
binascii.a2b_base64('anBqYW5lY2VrQGdtYWlsLmNvbQ==\n')
------------------

"""

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import subprocess

pkgconfig = subprocess.Popen("pkg-config --cflags webkit-1.0",
                                            stdout=subprocess.PIPE, shell=True)
pkgconfig.wait()
extra_compile_args = pkgconfig.stdout.read().split()

pkgconfig = subprocess.Popen("pkg-config --libs webkit-1.0",
                                            stdout=subprocess.PIPE, shell=True)
pkgconfig.wait()
extra_link_args = pkgconfig.stdout.read().split()

setup(
    name='python-jswebkit',
    version='0.0.3',
    description='WebKit/JavaScriptCore Python bindings',
    long_description ="""It's a cython wrapper for JSContextRef in pywebkitgtk
 that way it can call JS functions etc""",
    author='Jiahua Huang',
    author_email='jhuangjiahua@gmail.com',
    license='LGPL-3',
    url="http://code.google.com/p/gwrite",
    download_url="http://code.google.com/p/gwrite",
        
    cmdclass = {'build_ext': build_ext},
    ext_modules = [Extension("jswebkit", ["jswebkit.pyx"],
                    extra_compile_args = extra_compile_args,
                    extra_link_args = extra_link_args

                    )]
)



cdef extern from "JavaScriptCore/JSBase.h":
    struct OpaqueJSContextGroup:
        pass
    ctypedef OpaqueJSContextGroup* JSContextGroupRef

    struct OpaqueJSContext:
        pass
    ctypedef OpaqueJSContext* JSContextRef
    ctypedef OpaqueJSContext* JSGlobalContextRef
    
    struct OpaqueJSString:
        pass
    ctypedef OpaqueJSString* JSStringRef
    
    struct OpaqueJSClass:
        pass
    ctypedef OpaqueJSClass* JSClassRef
    
    struct OpaqueJSPropertyNameArray:
        pass
    ctypedef OpaqueJSPropertyNameArray* JSPropertyNameArrayRef
    
    struct OpaqueJSPropertyNameAccumulator:
        pass
    ctypedef OpaqueJSPropertyNameAccumulator* JSPropertyNameAccumulatorRef
    
    struct OpaqueJSValue:
        pass
    ctypedef OpaqueJSValue* JSValueRef
    ctypedef OpaqueJSValue* JSObjectRef
    
    JSValueRef JSEvaluateScript(JSContextRef ctx, JSStringRef script, JSObjectRef thisObject, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
    bool JSCheckScriptSyntax(JSContextRef ctx, JSStringRef script, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
    void JSGarbageCollect(JSContextRef ctx)


"""
Copyright (C) 2009 john paul janecek, Free Beer (MIT) Licensing
Copyright (C) 2010 Jiahua Huang <jhuangjiahua@gmail.com>, MIT Licensing for this file

Comment by john paul janecek:
------------------
version 0.0003
Applied the patches supplied by MonkeeSage Thanx :)
This fixes the ucs2/ucs4 bug which I had earlier
Also MonkeeSage fixed other things, he make call function nicer and fixed lists so the work properly
Also fixed his patch so it works with both ucs2/ucs4 python
Fixed the setup.py so that it can use pkg-config instead

version 0.0002
I need a versioning system lol
cython wrapper for JSContextRef in pywebkitgtk
you need cython to make it
So you can call Javascript functions etc from python
Made by john paul janecek
Free Beer copyright, do what the heck you want with it, just give me credit
Also do not blame me if your things blow up
if you need to contact me, i might answer back :) I am lazy when it comes to making fixes
unless I actually am using library myself :)

my email
import binascii
binascii.a2b_base64('anBqYW5lY2VrQGdtYWlsLmNvbQ==\n')
------------------

"""
import sys
import types
cdef :
    ctypedef unsigned short bool
    
include "stdlib.pyi"
include "python.pyi"
include "jsbase.pyi"
include "jsstringref.pyi"
include "jsvalueref.pyi"
include "jsobjectref.pyi"

cdef object jsValueToPython(JSContextRef ctx,JSValueRef jsValue):
    cdef JSStringRef jsStr
    cdef int jsType = JSValueGetType(ctx,jsValue)
    cdef JSObject jsObject
    cdef JSFunction jsFunction
    cdef object result
    cdef int isFunction
    cdef bool bResult
    cdef size_t strlen
    if jsType == kJSTypeUndefined or jsType == kJSTypeNull :
        JSValueUnprotect(ctx,jsValue)
        return None
    elif jsType == kJSTypeBoolean :
        bResult = JSValueToBoolean(ctx,jsValue)
        JSValueUnprotect(ctx,jsValue)
        if bResult > 0 :
            return True
        else :
            return False
    elif jsType == kJSTypeNumber:
        result = JSValueToNumber(ctx,jsValue,NULL)
        JSValueUnprotect(ctx,jsValue)
        return result
    elif jsType == kJSTypeString :
        jsStr = JSValueToStringCopy(ctx,jsValue,NULL)
        strlen = JSStringGetLength(jsStr)
        strlen *= 2
        result = PyUnicode_DecodeUTF16(JSStringGetCharactersPtr(jsStr), strlen, NULL, 0)
        JSStringRelease(jsStr)
        JSValueUnprotect(ctx,jsValue)
        return result
    else :
        if JSObjectIsFunction(ctx,jsValue) > 0 :
            jsFunction = JSFunction()
            jsFunction.setup(ctx,jsValue)
            return jsFunction
        else :
            jsObject = JSObject()
            jsObject.setup(ctx,jsValue)
            return jsObject
    return None

class JSException(Exception):
    def __init__(self,name,message):
        self.name = name
        self.mess = message
    
    def __str__(self):
        return "JSException name : %s message : %s" % (self.name,self.mess)
        
cdef object makeException(JSContextRef ctx,JSValueRef jsException) :
    e = jsValueToPython(ctx,jsException)
    return JSException(e.name,e.message)

cdef object JSStringRefToPython(JSStringRef jsString) :
    cdef size_t strlen = JSStringGetLength(jsString)
    strlen *= 2
    result = PyUnicode_DecodeUTF16(JSStringGetCharactersPtr(jsString), strlen, NULL, 0)
    return result

cdef JSStringRef PythonToJSString(object pyStr) :
    """result has to be released"""
    pyStr = unicode(pyStr).encode("utf-8")
    cdef JSStringRef jsStr = JSStringCreateWithUTF8CString(PyString_AsString(pyStr))
    return jsStr

cdef JSValueRef PythonTojsValue(JSContextRef ctx,object pyValue):
    if isinstance(pyValue,types.NoneType):
        return JSValueMakeNull(ctx)
    elif isinstance(pyValue,types.BooleanType):
        return JSValueMakeBoolean(ctx,pyValue)
    elif isinstance(pyValue,(types.IntType,types.FloatType)) :
        return JSValueMakeNumber(ctx,pyValue)
    elif isinstance(pyValue,types.StringTypes) :
        return JSValueMakeString(ctx,PythonToJSString(pyValue))
    elif isinstance(pyValue,JSObject):
        return (<JSObject>pyValue).jsObject
    else :
        raise ValueError

cdef class JSObject:
    cdef JSContextRef ctx
    cdef JSObjectRef jsObject
    cdef object propertyNames
    cdef int index
    
    def __init__(self) :
        self.ctx = NULL
        self.jsObject = NULL
        self.index = 0
        
    cdef setup(self,JSContextRef ctx,JSObjectRef jsObject):
        self.ctx = ctx
        self.jsObject = jsObject
        self.propertyNames = dict.fromkeys(self.getPropertyNames(),True)
        
    def getPropertyNames(self):
        cdef JSPropertyNameArrayRef nameArray = JSObjectCopyPropertyNames(self.ctx,self.jsObject)
        names = []
        for i in range(JSPropertyNameArrayGetCount(nameArray)):
            names.append(JSStringRefToPython(JSPropertyNameArrayGetNameAtIndex(nameArray,i)))
        JSPropertyNameArrayRelease(nameArray)
        return names
    
    def __getattr__(self,name):
        cdef JSStringRef jsStr
        cdef JSValueRef jsException
        try :            
            self.propertyNames[name]
            jsStr = JSStringCreateWithUTF8CString(name) #has to be a UTF8 string
            result = jsValueToPython(self.ctx,JSObjectGetProperty(self.ctx,self.jsObject,jsStr,NULL))
            JSStringRelease(jsStr)
            return result
        except KeyError :
            raise AttributeError,name
         
    def __setattr__(self,name,value):
        cdef JSStringRef jsStr 
        self.propertyNames[name] = True
        jsStr = JSStringCreateWithUTF8CString(name) #has to be a UTF8 string
        JSObjectSetProperty(self.ctx,self.jsObject,jsStr,PythonTojsValue(self.ctx,value),kJSPropertyAttributeNone,NULL)
        JSStringRelease(jsStr)
        
    def __getitem__(self,key) :
        cdef JSValueRef jsValueRef
        cdef JSValueRef jsException = NULL
        jsValueRef = JSObjectGetPropertyAtIndex(self.ctx,self.jsObject,key,&jsException)
        if jsException != NULL :
            raise makeException(self.ctx,jsException)
        return jsValueToPython(self.ctx,jsValueRef)
        
    def __setitem__(self,key,value):
        cdef JSValueRef jsValue = PythonTojsValue(self.ctx,value)
        cdef JSValueRef jsException = NULL
        JSObjectSetPropertyAtIndex(self.ctx, self.jsObject, key, jsValue, &jsException)
        if jsException != NULL :
            raise makeException(self.ctx,jsException)
            
    def __del__(self):
        JSValueUnprotect(self.ctx,self.jsObject)
    
    # these are container methods, so that lists behave correctly
    def __iter__(self):
        return self
        
    def __next__(self): # 2.6 and 3.0 now use __next__
        if self.index < self.length:
            value = self[self.index]
            self.index += 1
            return value
        else:
            self.index = 0
            raise StopIteration
        
    def next(self): # wrapper for backwards compatibility
        return self.__next__()

    def __len__(self):
        length = len(self.getPropertyNames())
        if hasattr(self, "length"):
            length += int(self.length)
        return length


cdef class JSFunction(JSObject):
    def __init__(self) :
        JSObject.__init__(self)
        
    def __call__(self,thisObj,*args):
        cdef JSValueRef *jsArgs
        cdef JSValueRef result
        cdef JSObjectRef jsThisObject
        cdef JSValueRef jsError = NULL
        if len(args) :
            jsArgs = <JSValueRef *>malloc(len(args) * sizeof(JSValueRef))
        else :
            jsArgs = NULL
        for i,arg in enumerate(args) :
            jsArgs[i] = PythonTojsValue(self.ctx,arg)
        if thisObj :
            jsThisObject = (<JSObject>thisObj).jsObject
        else :
            jsThisObject = NULL
        result = JSObjectCallAsFunction(self.ctx,self.jsObject,jsThisObject,len(args),jsArgs,&jsError)
        free(jsArgs)
        if jsError!=NULL :
            raise makeException(self.ctx,jsError)
        return jsValueToPython(self.ctx,result)
 

cdef class JSContext:
    cdef JSContextRef jsCtx
    cdef object ctx
    def __init__(self,ctx):
        self.ctx = ctx
        self.jsCtx = <JSContextRef>PyCObject_AsVoidPtr(ctx)
        
    def EvaluateScript(self,script,thisObject = None ,sourceURL = None,startingLineNumber = 1):
        script = unicode(script).encode("utf-8")
        cdef JSStringRef jsScript = JSStringCreateWithUTF8CString(PyString_AsString(script))
        cdef JSValueRef jsException = NULL
        cdef JSValueRef jsValue = JSEvaluateScript(self.jsCtx,jsScript,<JSObjectRef>NULL,<JSStringRef>NULL,startingLineNumber,&jsException)        
        JSValueProtect(self.jsCtx,jsValue)
        JSStringRelease(jsScript)
        if jsException!=NULL :
            raise makeException(self.jsCtx,jsException)
        return jsValueToPython(self.jsCtx,jsValue)
    
    def getCtx(self):
        return self.ctx
    
    
